import cirq
import numpy as np
import pytest

import stimcirq


def test_end_to_end():
    sampler = stimcirq.StimSampler()
    a, b = cirq.LineQubit.range(2)
    result = sampler.run(
        cirq.Circuit(
            cirq.H(a),
            cirq.CNOT(a, b),
            cirq.X(a) ** 0.5,
            cirq.X(b) ** 0.5,
            cirq.measure(a, key='a'),
            cirq.measure(b, key='b'),
        ),
        repetitions=1000,
    )
    np.testing.assert_array_equal(result.measurements['a'], result.measurements['b'] ^ 1)


def test_endian():
    a, b = cirq.LineQubit.range(2)
    circuit = cirq.Circuit(cirq.X(a), cirq.measure(a, b, key='out'))
    s1 = cirq.Simulator().sample(circuit)
    s2 = stimcirq.StimSampler().sample(circuit)
    assert s1['out'][0] == s2['out'][0]


def test_custom_gates():
    s = stimcirq.StimSampler()
    a, b, c, d = cirq.LineQubit.range(4)

    class GoodGate(cirq.SingleQubitGate):
        def _num_qubits_(self) -> int:
            return 1

        def _decompose_(self, qubits):
            return [cirq.X.on_each(*qubits)]

    class IndirectlyGoodGate(cirq.SingleQubitGate):
        def _decompose_(self, qubits):
            return [GoodGate().on_each(*qubits)]

    class BadGate(cirq.SingleQubitGate):
        pass

    class IndirectlyBadGate(cirq.SingleQubitGate):
        def _decompose_(self):
            return [BadGate()]

    np.testing.assert_array_equal(
        s.sample(cirq.Circuit(GoodGate().on(a), cirq.measure(a, key='out')))['out'], [True]
    )

    np.testing.assert_array_equal(
        s.sample(cirq.Circuit(IndirectlyGoodGate().on(a), cirq.measure(a, key='out')))['out'],
        [True],
    )

    with pytest.raises(TypeError, match="BadGate.+into stim"):
        s.sample(cirq.Circuit(BadGate().on(a)))

    with pytest.raises(TypeError, match="IndirectlyBadGate.+into stim"):
        s.sample(cirq.Circuit(IndirectlyBadGate().on(a)))

    with pytest.raises(TypeError, match="into stim"):
        s.sample(cirq.Circuit(cirq.TOFFOLI(a, b, c)))


def test_custom_measurement():
    class XBasisMeasurement(cirq.Gate):
        def __init__(self, key: str):
            self.key = key

        def _num_qubits_(self) -> int:
            return 1

        def _decompose_(self, qubits):
            q, = qubits
            return [cirq.H(q), cirq.measure(q, key=self.key), cirq.H(q)]

    s = stimcirq.StimSampler()
    a, b = cirq.LineQubit.range(2)
    out = s.sample(cirq.Circuit(
        cirq.H(a),
        cirq.X(b),
        cirq.H(b),
        XBasisMeasurement("a").on(a),
        XBasisMeasurement("b").on(b),
    ))
    np.testing.assert_array_equal(out["a"], [0])
    np.testing.assert_array_equal(out["b"], [1])


def test_correlated_error():
    s = stimcirq.StimSampler()
    a, b = cirq.LineQubit.range(2)
    s.run(cirq.Circuit((cirq.X(a) * cirq.Y(b)).with_probability(0.1)))



from typing import List

import cirq

from ._cirq_to_stim import cirq_circuit_to_stim_data


class StimSampler(cirq.Sampler):
    """Samples stabilizer circuits using Stim.

    Supports circuits that contain Clifford operations, measurement operations, reset operations, and noise operations
    that can be decomposed into probabilistic Pauli operations. Unknown operations are supported as long as they provide
    a decomposition into supported operations via `cirq.decompose` (i.e. via a `_decompose_` method).

    Note that batch sampling is significantly faster (as in potentially thousands of times faster) than individual
    sampling, because it amortizes the cost of parsing and analyzing the circuit.
    """

    def run_sweep(
        self,
        program: cirq.Circuit,
        params: cirq.Sweepable,
        repetitions: int = 1,
    ) -> List[cirq.Result]:
        trial_results: List[cirq.Result] = []
        for param_resolver in cirq.to_resolvers(params):
            # Request samples from stim.
            instance = cirq.resolve_parameters(program, param_resolver)
            converted_circuit, key_ranges = cirq_circuit_to_stim_data(instance)
            samples = converted_circuit.compile_sampler().sample(repetitions)

            # Convert unlabelled samples into keyed results.
            k = 0
            measurements = {}
            for key, length in key_ranges:
                p = k
                k += length
                measurements[key] = samples[:, p:k]
            trial_results.append(cirq.Result(params=param_resolver, measurements=measurements))

        return trial_results



from typing import Callable, cast, Dict, Iterable, List, Optional, Sequence, Tuple, Type

import functools
import itertools
import math

import cirq
import stim


def cirq_circuit_to_stim_circuit(circuit: cirq.Circuit,
                                 *,
                                 qubit_to_index_dict: Optional[Dict[cirq.Qid, int]] = None) -> stim.Circuit:
    """Converts a cirq circuit into an equivalent stim circuit.

    Not all circuits can be converted. In order for a circuit to be convertible, all of its operations must be
    convertible.

    An operation is convertible if:
        - It is a stabilizer gate or probabilistic Pauli gate from cirq
            - cirq.H
            - cirq.S
            - cirq.X
            - cirq.X**0.5
            - cirq.CNOT
            - cirq.ResetChannel()
            - cirq.X.with_probability(p)
            - cirq.DepolarizingChannel(p, n_qubits=1 or 2)
            - etc
        - Or it has a _decompose_ method that yields convertible operations.
        - Or it has a correctly implemented _stim_conversion_ method.

    Args:
        circuit: The circuit to convert.
        qubit_to_index_dict: Optional. Which integer each qubit should get mapped to. If not specified, defaults to
            indexing qubits in the circuit in sorted order.

    Returns:
        The converted circuit.

    Examples:
        >>> import cirq, stimcirq
        >>> a = cirq.NamedQubit("zero")
        >>> b = cirq.NamedQubit("two")
        >>> stimcirq.cirq_circuit_to_stim_circuit(cirq.Circuit(
        ...     cirq.Moment(cirq.H(a)),
        ...     cirq.Moment(cirq.CNOT(a, b)),
        ...     cirq.Moment(
        ...         cirq.X(a).with_probability(0.25),
        ...         cirq.Z(b).with_probability(0.25),
        ...     ),
        ...     cirq.Moment(),
        ...     cirq.Moment(),
        ...     cirq.Moment(cirq.DepolarizingChannel(0.125, n_qubits=2).on(b, a)),
        ...     cirq.Moment(cirq.measure(a, b)),
        ... ), qubit_to_index_dict={a: 0, b: 2})
        stim.Circuit('''
            H 0
            TICK
            CX 0 2
            TICK
            X_ERROR(0.25) 0
            Z_ERROR(0.25) 2
            TICK
            TICK
            TICK
            DEPOLARIZE2(0.125) 2 0
            TICK
            M 0 2
            TICK
        ''')

    Here is an example of a _stim_conversion_ method:

        def _stim_conversion_(
                self,

                # The stim circuit being built. Add onto it.
                edit_circuit: stim.Circuit,

                # Metadata about measurement groupings needed by stimcirq.StimSampler.
                # If your gate contains a measurement, it has to append how many qubits
                # that measurement measures (and its key) into this list.
                edit_measurement_key_lengths: List[Tuple[str, int]],

                # The indices of qubits the gate is operating on.
                targets: List[int],

                # Forward compatibility with future arguments.
                **kwargs):

            edit_circuit.append_operation("H", targets)
    """
    return cirq_circuit_to_stim_data(circuit, q2i=qubit_to_index_dict)[0]


def cirq_circuit_to_stim_data(
    circuit: cirq.Circuit, *, q2i: Optional[Dict[cirq.Qid, int]] = None
) -> Tuple[stim.Circuit, List[Tuple[str, int]]]:
    """Converts a Cirq circuit into a Stim circuit and also metadata about where measurements go."""
    if q2i is None:
        q2i = {q: i for i, q in enumerate(sorted(circuit.all_qubits()))}
    out = stim.Circuit()
    key_out: List[Tuple[str, int]] = []

    for q in sorted(circuit.all_qubits()):
        if isinstance(q, cirq.LineQubit):
            i = q2i[q]
            if i != q.x:
                out.append_operation("QUBIT_COORDS", [i], [q.x])
        elif isinstance(q, cirq.GridQubit):
            out.append_operation("QUBIT_COORDS", [q2i[q]], [q.row, q.col])

    for moment in circuit:
        _c2s_helper(moment, q2i, out, key_out)
        out.append_operation("TICK", [])
    return out, key_out


StimTypeHandler = Callable[[stim.Circuit, cirq.Gate, List[int]], None]


@functools.lru_cache(maxsize=1)
def gate_to_stim_append_func() -> Dict[cirq.Gate, Callable[[stim.Circuit, List[int]], None]]:
    """A dictionary mapping specific gate instances to stim circuit appending functions."""
    x = (cirq.X, False)
    y = (cirq.Y, False)
    z = (cirq.Z, False)
    nx = (cirq.X, True)
    ny = (cirq.Y, True)
    nz = (cirq.Z, True)

    def do_nothing(c, t):
        pass

    def use(
        *gates: str, individuals: Sequence[Tuple[str, int]] = ()
    ) -> Callable[[stim.Circuit, List[int]], None]:
        if len(gates) == 1 and not individuals:
            (g,) = gates
            return lambda c, t: c.append_operation(g, t)

        if not individuals:

            def do(c, t):
                for g in gates:
                    c.append_operation(g, t)

        else:

            def do(c, t):
                for g in gates:
                    c.append_operation(g, t)
                for g, k in individuals:
                    c.append_operation(g, [t[k]])

        return do

    sqcg = cirq.SingleQubitCliffordGate.from_xz_map
    paulis = cast(List[cirq.Pauli], [cirq.X, cirq.Y, cirq.Z])

    return {
        cirq.ResetChannel(): use("R"),
        # Identities.
        cirq.I: use("I"),
        cirq.H ** 0: do_nothing,
        cirq.X ** 0: do_nothing,
        cirq.Y ** 0: do_nothing,
        cirq.Z ** 0: do_nothing,
        cirq.ISWAP ** 0: do_nothing,
        cirq.SWAP ** 0: do_nothing,
        # Common named gates.
        cirq.H: use("H"),
        cirq.X: use("X"),
        cirq.Y: use("Y"),
        cirq.Z: use("Z"),
        cirq.X ** 0.5: use("SQRT_X"),
        cirq.X ** -0.5: use("SQRT_X_DAG"),
        cirq.Y ** 0.5: use("SQRT_Y"),
        cirq.Y ** -0.5: use("SQRT_Y_DAG"),
        cirq.Z ** 0.5: use("SQRT_Z"),
        cirq.Z ** -0.5: use("SQRT_Z_DAG"),
        cirq.CNOT: use("CNOT"),
        cirq.CZ: use("CZ"),
        cirq.ISWAP: use("ISWAP"),
        cirq.ISWAP ** -1: use("ISWAP_DAG"),
        cirq.ISWAP ** 2: use("Z"),
        cirq.SWAP: use("SWAP"),
        cirq.X.controlled(1): use("CX"),
        cirq.Y.controlled(1): use("CY"),
        cirq.Z.controlled(1): use("CZ"),
        cirq.XX**0.5: use("SQRT_XX"),
        cirq.YY**0.5: use("SQRT_YY"),
        cirq.ZZ**0.5: use("SQRT_ZZ"),
        cirq.XX**-0.5: use("SQRT_XX_DAG"),
        cirq.YY**-0.5: use("SQRT_YY_DAG"),
        cirq.ZZ**-0.5: use("SQRT_ZZ_DAG"),
        # All 24 cirq.SingleQubitCliffordGate instances.
        sqcg(x, y): use("SQRT_X_DAG"),
        sqcg(x, ny): use("SQRT_X"),
        sqcg(nx, y): use("H_YZ"),
        sqcg(nx, ny): use("H_YZ", "X"),
        sqcg(x, z): do_nothing,
        sqcg(x, nz): use("X"),
        sqcg(nx, z): use("Z"),
        sqcg(nx, nz): use("Y"),
        sqcg(y, x): use("C_XYZ"),
        sqcg(y, nx): use("S", "SQRT_Y_DAG"),
        sqcg(ny, x): use("S_DAG", "SQRT_Y"),
        sqcg(ny, nx): use("S_DAG", "SQRT_Y_DAG"),
        sqcg(y, z): use("S"),
        sqcg(y, nz): use("H_XY"),
        sqcg(ny, z): use("S_DAG"),
        sqcg(ny, nz): use("H_XY", "Z"),
        sqcg(z, x): use("H"),
        sqcg(z, nx): use("SQRT_Y_DAG"),
        sqcg(nz, x): use("SQRT_Y"),
        sqcg(nz, nx): use("H", "Y"),
        sqcg(z, y): use("C_ZYX"),
        sqcg(z, ny): use("SQRT_Y_DAG", "S"),
        sqcg(nz, y): use("SQRT_Y", "S"),
        sqcg(nz, ny): use("SQRT_Y", "S_DAG"),
        # All 36 cirq.PauliInteractionGate instances.
        **{
            cirq.PauliInteractionGate(p0, s0, p1, s1): use(
                f"{p0}C{p1}", individuals=[(str(p1), 1)] * s0 + [(str(p0), 0)] * s1
            )
            for p0, s0, p1, s1 in itertools.product(paulis, [False, True], repeat=2)
        },
    }


@functools.lru_cache()
def gate_type_to_stim_append_func() -> Dict[Type[cirq.Gate], StimTypeHandler]:
    """A dictionary mapping specific gate types to stim circuit appending functions."""
    return {
        cirq.ControlledGate: cast(StimTypeHandler, _stim_append_controlled_gate),
        cirq.DensePauliString: cast(StimTypeHandler, _stim_append_dense_pauli_string_gate),
        cirq.MutableDensePauliString: cast(StimTypeHandler, _stim_append_dense_pauli_string_gate),
        cirq.AsymmetricDepolarizingChannel: cast(StimTypeHandler, _stim_append_asymmetric_depolarizing_channel),
        cirq.BitFlipChannel: lambda c, g, t: c.append_operation(
            "X_ERROR", t, cast(cirq.BitFlipChannel, g).p
        ),
        cirq.PhaseFlipChannel: lambda c, g, t: c.append_operation(
            "Z_ERROR", t, cast(cirq.PhaseFlipChannel, g).p
        ),
        cirq.PhaseDampingChannel: lambda c, g, t: c.append_operation(
            "Z_ERROR", t, 0.5 - math.sqrt(1 - cast(cirq.PhaseDampingChannel, g).gamma) / 2
        ),
        cirq.RandomGateChannel: cast(StimTypeHandler, _stim_append_random_gate_channel),
        cirq.DepolarizingChannel: cast(StimTypeHandler, _stim_append_depolarizing_channel),
    }


def _stim_append_measurement_gate(circuit: stim.Circuit, gate: cirq.MeasurementGate, targets: List[int]):
    for i, b in enumerate(gate.invert_mask):
        if b:
            targets[i] = stim.target_inv(targets[i])
    circuit.append_operation("M",  targets)


def _stim_append_dense_pauli_string_gate(c: stim.Circuit, g: cirq.BaseDensePauliString, t: List[int]):
    gates = [None, "X", "Y", "Z"]
    for p, k in zip(g.pauli_mask, t):
        if p:
            c.append_operation(gates[p], [k])


def _stim_append_asymmetric_depolarizing_channel(c: stim.Circuit, g: cirq.AsymmetricDepolarizingChannel, t: List[int]):
    c.append_operation("PAULI_CHANNEL_1", t, [g.p_x, g.p_y, g.p_z])


def _stim_append_depolarizing_channel(c: stim.Circuit, g: cirq.DepolarizingChannel, t: List[int]):
    if g.num_qubits() == 1:
        c.append_operation("DEPOLARIZE1", t, g.p)
    elif g.num_qubits() == 2:
        c.append_operation("DEPOLARIZE2", t, g.p)
    else:
        raise TypeError(f"Don't know how to turn {g!r} into Stim operations.")


def _stim_append_controlled_gate(c: stim.Circuit, g: cirq.ControlledGate, t: List[int]):
    if isinstance(g.sub_gate, cirq.BaseDensePauliString) and g.num_controls() == 1:
        gates = [None, "CX", "CY", "CZ"]
        for p, k in zip(g.sub_gate.pauli_mask, t[1:]):
            if p:
                c.append_operation(gates[p], [t[0], k])
        if g.sub_gate.coefficient == 1j:
            c.append_operation("S", t[:1])
        elif g.sub_gate.coefficient == -1:
            c.append_operation("Z", t[:1])
        elif g.sub_gate.coefficient == -1j:
            c.append_operation("S_DAG", t[:1])
        elif g.sub_gate.coefficient == 1:
            pass
        else:
            raise TypeError(f"Phase kickback from {g!r} isn't a stabilizer operation.")
        return

    raise TypeError(f"Don't know how to turn controlled gate {g!r} into Stim operations.")


def _stim_append_random_gate_channel(c: stim.Circuit, g: cirq.RandomGateChannel, t: List[int]):
    if g.sub_gate in [cirq.X, cirq.Y, cirq.Z]:
        c.append_operation(f"{g.sub_gate}_ERROR", t, g.probability)
    elif isinstance(g.sub_gate, cirq.DensePauliString):
        target_p = [None, stim.target_x, stim.target_y, stim.target_z]
        pauli_targets = [
            target_p[p](t)
            for t, p in zip(t, g.sub_gate.pauli_mask)
        ]
        c.append_operation(f"CORRELATED_ERROR", pauli_targets, g.probability)
    else:
        raise NotImplementedError(f"Don't know how to turn probabilistic {g!r} into Stim operations.")


def _c2s_helper(
    operations: Iterable[cirq.Operation], q2i: Dict[cirq.Qid, int], out: stim.Circuit,
    key_out: List[Tuple[str, int]]
):
    g2f = gate_to_stim_append_func()
    t2f = gate_type_to_stim_append_func()
    for op in operations:
        gate = op.gate
        targets = [q2i[q] for q in op.qubits]

        custom_method = getattr(op, '_stim_conversion_', getattr(gate, '_stim_conversion_', None))
        if custom_method is not None:
            custom_method(
                dont_forget_your_star_star_kwargs=True,
                edit_circuit=out,
                edit_measurement_key_lengths=key_out,
                targets=targets)
            continue

        # Special case measurement, because of its metadata.
        if isinstance(gate, cirq.MeasurementGate):
            key_out.append((gate.key, len(targets)))
            _stim_append_measurement_gate(out, gate, targets)
            continue

        # Look for recognized gate values like cirq.H.
        val_append_func = g2f.get(gate)
        if val_append_func is not None:
            val_append_func(out, targets)
            continue

        # Look for recognized gate types like cirq.DepolarizingChannel.
        type_append_func = t2f.get(type(gate))
        if type_append_func is not None:
            type_append_func(out, gate, targets)
            continue

        # Ask unrecognized operations to decompose themselves into simpler operations.
        try:
            _c2s_helper(cirq.decompose_once(op), q2i, out, key_out)
        except TypeError as ex:
            raise TypeError(f"Don't know how to translate {op!r} into stim gates.") from ex

    return out



import inspect
from typing import Tuple, Union, Callable, Any, cast

import cirq
import pytest
import stim

import stimcirq
from ._stim_to_cirq import stim_to_cirq_gate_table, not_handled_or_handled_specially_set


def test_two_qubit_asymmetric_depolarizing_channel():
    r = stimcirq.TwoQubitAsymmetricDepolarizingChannel([0.125, 0, 0, 0, 0, 0, 0.375, 0, 0, 0, 0, 0, 0, 0.25, 0])
    assert r._dense_mixture_() == [
        (0.25, cirq.DensePauliString("II")),
        (0.125, cirq.DensePauliString("IX")),
        (0.375, cirq.DensePauliString("XZ")),
        (0.25, cirq.DensePauliString("ZY")),
    ]
    cirq.testing.assert_has_diagram(cirq.Circuit(r.on(*cirq.LineQubit.range(2))), """
0: ───PauliMix(II:0.25,IX:0.125,XZ:0.375,ZY:0.25)───
      │
1: ───#2────────────────────────────────────────────
    """)
    assert eval(repr(r), {'stimcirq': stimcirq}) == r


def test_stim_circuit_to_cirq_circuit():
    circuit = stimcirq.stim_circuit_to_cirq_circuit(stim.Circuit("""
        X 0
        CNOT 0 1
        X_ERROR(0.125) 0 1
        CORRELATED_ERROR(0.25) X0 Y1 Z2
        M 1
        M !1
        TICK
        MR 0 !1
    """))
    a, b, c = cirq.LineQubit.range(3)
    assert circuit == cirq.Circuit(
        cirq.Moment(cirq.X(a)),
        cirq.Moment(cirq.CNOT(a, b)),
        cirq.Moment(
            cirq.X.with_probability(0.125).on(a),
            cirq.X.with_probability(0.125).on(b),
        ),
        cirq.Moment(
            cirq.PauliString({a: cirq.X, b: cirq.Y, c: cirq.Z}).with_probability(0.25),
        ),
        cirq.Moment(
            cirq.measure(b, key="0"),
        ),
        cirq.Moment(
            cirq.measure(b, key="1", invert_mask=(True,)),
        ),
        cirq.Moment(
            stimcirq.MeasureAndOrReset(measure=True, reset=True, basis='Z', invert_measure=False, key='2').on(a),
            stimcirq.MeasureAndOrReset(measure=True, reset=True, basis='Z', invert_measure=True, key='3').on(b),
        ),
    )


@pytest.mark.parametrize("key,handler", list(stim_to_cirq_gate_table().items()) + [
    ("m-case", stimcirq.MeasureAndOrReset(
        measure=True,
        reset=reset,
        basis=basis,
        invert_measure=invert,
        key='0'
    ))
    for reset in [False, True]
    for basis in ['X', 'Y', 'Z']
    for invert in [False, True]
    if reset or basis != 'Z'
] + [
    ("+M", cirq.MeasurementGate(num_qubits=1, key='0', invert_mask=())),
    ("-M", cirq.MeasurementGate(num_qubits=1, key='0', invert_mask=(True,))),
])
def test_exact_gate_round_trips(key: str, handler: Union[cirq.Gate, Callable[[Any], cirq.Gate], Tuple]):
    if handler == ():
        return
    if isinstance(handler, cirq.Gate):
        gate = handler
    else:
        try:
            gate = handler(0.125)
        except NotImplementedError:
            assert key in not_handled_or_handled_specially_set()
            return
        except:
            try:
                gate = handler([k/128 for k in range(1, 4)])
            except:
                gate = handler([k/128 for k in range(1, 16)])
    n = cirq.num_qubits(gate)
    qs = cirq.LineQubit.range(n)
    original = cirq.Circuit(gate.on(*qs))
    converted = stimcirq.cirq_circuit_to_stim_circuit(original)
    restored = stimcirq.stim_circuit_to_cirq_circuit(converted)
    assert original == restored


def test_round_trip_preserves_moment_structure():
    a, b = cirq.LineQubit.range(2)
    circuit = cirq.Circuit(
        cirq.Moment(),
        cirq.Moment(cirq.H(a)),
        cirq.Moment(cirq.H(b)),
        cirq.Moment(),
        cirq.Moment(),
        cirq.Moment(cirq.CNOT(a, b)),
        cirq.Moment(),
    )
    converted = stimcirq.cirq_circuit_to_stim_circuit(circuit)
    restored = stimcirq.stim_circuit_to_cirq_circuit(converted)
    assert restored == circuit


def test_circuit_diagram():
    cirq.testing.assert_has_diagram(
        stimcirq.stim_circuit_to_cirq_circuit(stim.Circuit("""
            M 0
            MX 0
            MY 0
            MZ 0
            R 0
            RX 0
            RY 0
            RZ 0
            MR 0
            MRX 0
            MRY 0
            MRZ 0
        """)), """
0: ───M───MX('1')───MY('2')───M('3')───R───RX───RY───R───MR('4')───MRX('5')───MRY('6')───MR('7')───
        """)


def test_all_known_gates_explicitly_handled():
    output_from_stim_help_gates = """
CNOT
CORRELATED_ERROR
CX
CY
CZ
C_XYZ
C_ZYX
DEPOLARIZE1
DEPOLARIZE2
DETECTOR
E
ELSE_CORRELATED_ERROR
H
H_XY
H_XZ
H_YZ
I
ISWAP
ISWAP_DAG
M
MR
MRX
MRY
MRZ
MX
MY
MZ
OBSERVABLE_INCLUDE
PAULI_CHANNEL_1
PAULI_CHANNEL_2
QUBIT_COORDS
R
REPEAT
RX
RY
RZ
S
SHIFT_COORDS
SQRT_X
SQRT_XX
SQRT_XX_DAG
SQRT_X_DAG
SQRT_Y
SQRT_YY
SQRT_YY_DAG
SQRT_Y_DAG
SQRT_Z
SQRT_ZZ
SQRT_ZZ_DAG
SQRT_Z_DAG
SWAP
S_DAG
TICK
X
XCX
XCY
XCZ
X_ERROR
Y
YCX
YCY
YCZ
Y_ERROR
Z
ZCX
ZCY
ZCZ
Z_ERROR
    """
    gates = output_from_stim_help_gates.strip().split()
    handled = stim_to_cirq_gate_table().keys() | not_handled_or_handled_specially_set()
    for gate_name in gates:
        assert gate_name in handled, gate_name


def test_line_grid_qubit_round_trip():
    c = cirq.Circuit(
        cirq.H(cirq.LineQubit(101)),
        cirq.Z(cirq.LineQubit(200.5)),
        cirq.X(cirq.GridQubit(2, 3.5)),
        cirq.Y(cirq.GridQubit(-2, 5)),
    )
    s = stimcirq.cirq_circuit_to_stim_circuit(c)
    assert s == stim.Circuit("""
        QUBIT_COORDS(-2, 5) 0
        QUBIT_COORDS(2, 3.5) 1
        QUBIT_COORDS(101) 2
        QUBIT_COORDS(200.5) 3
        H 2
        Z 3
        X 1
        Y 0
        TICK
    """)
    c2 = stimcirq.stim_circuit_to_cirq_circuit(s)
    for q in c2.all_qubits():
        if q == cirq.LineQubit(101):
            assert isinstance(cast(cirq.LineQubit, q).x, int)
        if q == cirq.GridQubit(2, 3.5):
            assert isinstance(cast(cirq.GridQubit, q).row, int)
    assert c2 == c

    assert stimcirq.stim_circuit_to_cirq_circuit(stim.Circuit("""
        QUBIT_COORDS(10) 0
        SHIFT_COORDS(1, 2, 3)
        QUBIT_COORDS(20, 30) 1
        H 0 1
    """)) == cirq.Circuit(cirq.H(cirq.LineQubit(10)), cirq.H(cirq.GridQubit(21, 32)))


def test_noisy_measurements():
    s = stim.Circuit("""
        MX(0.125) 0
        MY(0.125) 1
        MZ(0.125) 2
        MRX(0.125) 3
        MRY(0.125) 4
        MRZ(0.25) 5
        TICK
    """)
    c = stimcirq.stim_circuit_to_cirq_circuit(s)
    assert c == cirq.Circuit(
        stimcirq.MeasureAndOrReset(measure=True, reset=False, basis='X', invert_measure=False, key='0', measure_flip_probability=0.125).on(cirq.LineQubit(0)),
        stimcirq.MeasureAndOrReset(measure=True, reset=False, basis='Y', invert_measure=False, key='1', measure_flip_probability=0.125).on(cirq.LineQubit(1)),
        stimcirq.MeasureAndOrReset(measure=True, reset=False, basis='Z', invert_measure=False, key='2', measure_flip_probability=0.125).on(cirq.LineQubit(2)),
        stimcirq.MeasureAndOrReset(measure=True, reset=True, basis='X', invert_measure=False, key='3', measure_flip_probability=0.125).on(cirq.LineQubit(3)),
        stimcirq.MeasureAndOrReset(measure=True, reset=True, basis='Y', invert_measure=False, key='4', measure_flip_probability=0.125).on(cirq.LineQubit(4)),
        stimcirq.MeasureAndOrReset(measure=True, reset=True, basis='Z', invert_measure=False, key='5', measure_flip_probability=0.25).on(cirq.LineQubit(5)),
    )
    assert stimcirq.cirq_circuit_to_stim_circuit(c) == s
    cirq.testing.assert_equivalent_repr(c, global_vals={'stimcirq': stimcirq})



from ._cirq_to_stim import cirq_circuit_to_stim_circuit
from ._stim_sampler import StimSampler
from ._stim_to_cirq import stim_circuit_to_cirq_circuit, MeasureAndOrReset, TwoQubitAsymmetricDepolarizingChannel

# Workaround for doctest not searching imported objects.
__test__ = {
    "StimSampler": StimSampler,
    "cirq_circuit_to_stim_circuit": cirq_circuit_to_stim_circuit,
    "stim_circuit_to_cirq_circuit": stim_circuit_to_cirq_circuit,
    "MeasureAndOrReset": MeasureAndOrReset,
    "TwoQubitAsymmetricDepolarizingChannel": TwoQubitAsymmetricDepolarizingChannel,
}



import collections
import functools
from typing import Callable, Dict, List, Tuple, Union, Iterator, cast, Sequence, Iterable, Set, Any, DefaultDict

import cirq
import stim


@cirq.value_equality
class TwoQubitAsymmetricDepolarizingChannel(cirq.Gate):
    def __init__(self, probabilities: Sequence[float]):
        if len(probabilities) != 15:
            raise ValueError(len(probabilities) != 15)
        self.probabilities = tuple(probabilities)

    def _num_qubits_(self):
        return 2

    def _value_equality_values_(self):
        return self.probabilities

    def _has_mixture_(self):
        return True

    def _dense_mixture_(self):
        result = [(1 - sum(self.probabilities), cirq.DensePauliString([0, 0]))]
        result.extend([
            (p, cirq.DensePauliString([((k + 1) >> 2) & 3, (k + 1) & 3]))
            for k, p in enumerate(self.probabilities)
        ])
        return [(p, g) for p, g in result if p]

    def _mixture_(self):
        return [(p, cirq.unitary(g)) for p, g in self._dense_mixture_()]

    def _circuit_diagram_info_(self, args: cirq.CircuitDiagramInfoArgs):
        result = []
        for p, d in self._dense_mixture_():
            result.append(str(d)[1:] + ":" + args.format_real(p))
        return "PauliMix(" + ",".join(result) + ")", "#2"

    def _stim_conversion_(
            self,
            edit_circuit: stim.Circuit,
            targets: List[int],
            **kwargs):
        edit_circuit.append_operation("PAULI_CHANNEL_2", targets, self.probabilities)

    def __repr__(self):
        return f"stimcirq.TwoQubitAsymmetricDepolarizingChannel({self.probabilities!r})"


@cirq.value_equality
class MeasureAndOrReset(cirq.SingleQubitGate):
    def __init__(self, measure: bool, reset: bool, basis: str, invert_measure: bool, key: str, measure_flip_probability: float = 0):
        self.measure = measure
        self.reset = reset
        self.basis = basis
        self.invert_measure = invert_measure
        self.key = key
        self.measure_flip_probability = measure_flip_probability

    def _value_equality_values_(self):
        return self.measure, self.reset, self.basis, self.invert_measure, self.key, self.measure_flip_probability

    def _decompose_(self, qubits):
        q, = qubits
        if self.measure:
            if self.basis == 'X':
                yield cirq.H(q)
            elif self.basis == 'Y':
                yield cirq.X(q)**0.5
            if self.measure_flip_probability:
                raise NotImplementedError("Noisy measurement as a cirq operation.")
            else:
                yield cirq.measure(q, key=self.key, invert_mask=(True,) if self.invert_measure else ())
        if self.reset:
            yield cirq.ResetChannel().on(q)
        if self.measure or self.reset:
            if self.basis == 'X':
                yield cirq.H(q)
            elif self.basis == 'Y':
                yield cirq.X(q)**-0.5

    def with_key(self, key: str) -> 'MeasureAndOrReset':
        return MeasureAndOrReset(
            measure=self.measure,
            reset=self.reset,
            basis=self.basis,
            invert_measure=self.invert_measure,
            key=key,
            measure_flip_probability=self.measure_flip_probability
        )

    def with_bits_flipped(self, *bit_positions):
        assert bit_positions == (0,)
        return MeasureAndOrReset(
            measure=self.measure,
            reset=self.reset,
            basis=self.basis,
            invert_measure=not self.invert_measure,
            key=self.key,
            measure_flip_probability=self.measure_flip_probability
        )

    def _stim_op_name(self) -> str:
        result = ''
        if self.measure:
            result += "M"
        if self.reset:
            result += "R"
        if self.basis != 'Z':
            result += self.basis
        return result

    def _stim_conversion_(
            self,
            edit_circuit: stim.Circuit,
            targets: List[int],
            **kwargs):
        if self.invert_measure:
            targets[0] = stim.target_inv(targets[0])
        if self.measure_flip_probability:
            edit_circuit.append_operation(self._stim_op_name(), targets, self.measure_flip_probability)
        else:
            edit_circuit.append_operation(self._stim_op_name(), targets)

    def __str__(self) -> str:
        result = self._stim_op_name()
        if self.invert_measure:
            result = "!" + result
        if self.measure:
            result += f"('{self.key}')"
        if self.measure_flip_probability:
            result += f"^{self.measure_flip_probability:%}"
        return result

    def __repr__(self):
        return (f'stimcirq.MeasureAndOrReset('
                f'measure={self.measure!r}, '
                f'reset={self.reset!r}, '
                f'basis={self.basis!r}, '
                f'invert_measure={self.invert_measure!r}, '
                f'key={self.key!r}, '
                f'measure_flip_probability={self.measure_flip_probability!r})')


def args_to_cirq_depolarize_1(args: List[float]) -> cirq.AsymmetricDepolarizingChannel:
    if len(args) != 3:
        raise ValueError(f"len(args) != 3: {args!r}")
    return cirq.AsymmetricDepolarizingChannel(p_x=args[0], p_y=args[1], p_z=args[2])


def thrower(message) -> Callable[[Any], None]:
    def f(*args, **kwargs):
        raise NotImplementedError(message)
    return f


def _opt(args: Union[float, List[float]]) -> float:
    if not args:
        return 0
    return args if isinstance(args, float) else args[0]


@functools.lru_cache(maxsize=1)
def stim_to_cirq_gate_table() -> Dict[str, Union[Tuple, cirq.Gate, Callable[[Union[float, List[float]]], cirq.Gate]]]:
    return {
        "R": cirq.ResetChannel(),
        "RX": MeasureAndOrReset(measure=False, reset=True, basis='X', invert_measure=False, key='', measure_flip_probability=0),
        "RY": MeasureAndOrReset(measure=False, reset=True, basis='Y', invert_measure=False, key='', measure_flip_probability=0),
        "RZ": cirq.ResetChannel(),
        "M": lambda args: cirq.MeasurementGate(num_qubits=1, key='0') if not _opt(args) else MeasureAndOrReset(measure=True, reset=False, basis='Z', invert_measure=False, key='0', measure_flip_probability=_opt(args)),
        "MX": lambda args: MeasureAndOrReset(measure=True, reset=False, basis='X', invert_measure=False, key='0', measure_flip_probability=_opt(args)),
        "MY": lambda args: MeasureAndOrReset(measure=True, reset=False, basis='Y', invert_measure=False, key='0', measure_flip_probability=_opt(args)),
        "MZ": lambda args: cirq.MeasurementGate(num_qubits=1, key='0') if not _opt(args) else MeasureAndOrReset(measure=True, reset=False, basis='Z', invert_measure=False, key='0', measure_flip_probability=_opt(args)),
        "MR": lambda args: MeasureAndOrReset(measure=True, reset=True, basis='Z', invert_measure=False, key='0', measure_flip_probability=_opt(args)),
        "MRX": lambda args: MeasureAndOrReset(measure=True, reset=True, basis='X', invert_measure=False, key='0', measure_flip_probability=_opt(args)),
        "MRY": lambda args: MeasureAndOrReset(measure=True, reset=True, basis='Y', invert_measure=False, key='0', measure_flip_probability=_opt(args)),
        "MRZ": lambda args: MeasureAndOrReset(measure=True, reset=True, basis='Z', invert_measure=False, key='0', measure_flip_probability=_opt(args)),
        "I": cirq.I,
        "X": cirq.X,
        "Y": cirq.Y,
        "Z": cirq.Z,
        "H_XY": cirq.SingleQubitCliffordGate.from_xz_map(x_to=(cirq.Y, False), z_to=(cirq.Z, True)),
        "H": cirq.H,
        "H_XZ": cirq.H,
        "H_YZ": cirq.SingleQubitCliffordGate.from_xz_map(x_to=(cirq.X, True), z_to=(cirq.Y, False)),
        "SQRT_X": cirq.X**0.5,
        "SQRT_X_DAG": cirq.X**-0.5,
        "SQRT_Y": cirq.Y**0.5,
        "SQRT_Y_DAG": cirq.Y**-0.5,
        "C_XYZ": cirq.SingleQubitCliffordGate.from_xz_map(x_to=(cirq.Y, False), z_to=(cirq.X, False)),
        "C_ZYX": cirq.SingleQubitCliffordGate.from_xz_map(x_to=(cirq.Z, False), z_to=(cirq.Y, False)),
        "SQRT_XX": cirq.XX**0.5,
        "SQRT_YY": cirq.YY**0.5,
        "SQRT_ZZ": cirq.ZZ**0.5,
        "SQRT_XX_DAG": cirq.XX**-0.5,
        "SQRT_YY_DAG": cirq.YY**-0.5,
        "SQRT_ZZ_DAG": cirq.ZZ**-0.5,
        "S": cirq.S,
        "S_DAG": cirq.S**-1,
        "SQRT_Z": cirq.S,
        "SQRT_Z_DAG": cirq.S**-1,
        "SWAP": cirq.SWAP,
        "ISWAP": cirq.ISWAP,
        "ISWAP_DAG": cirq.ISWAP**-1,
        "XCX": cirq.PauliInteractionGate(cirq.X, False, cirq.X, False),
        "XCY": cirq.PauliInteractionGate(cirq.X, False, cirq.Y, False),
        "XCZ": cirq.PauliInteractionGate(cirq.X, False, cirq.Z, False),
        "YCX": cirq.PauliInteractionGate(cirq.Y, False, cirq.X, False),
        "YCY": cirq.PauliInteractionGate(cirq.Y, False, cirq.Y, False),
        "YCZ": cirq.PauliInteractionGate(cirq.Y, False, cirq.Z, False),
        "CX": cirq.CNOT,
        "CNOT": cirq.CNOT,
        "ZCX": cirq.CNOT,
        "CY": cirq.Y.controlled(1),
        "ZCY": cirq.Y.controlled(1),
        "CZ": cirq.CZ,
        "ZCZ": cirq.CZ,
        "DEPOLARIZE1": lambda arg: cirq.DepolarizingChannel(arg, 1),
        "DEPOLARIZE2": lambda arg: cirq.DepolarizingChannel(arg, 2),
        "X_ERROR": cirq.X.with_probability,
        "Y_ERROR": cirq.Y.with_probability,
        "Z_ERROR": cirq.Z.with_probability,
        "PAULI_CHANNEL_1": args_to_cirq_depolarize_1,
        "PAULI_CHANNEL_2": lambda args: TwoQubitAsymmetricDepolarizingChannel(args),
        "DETECTOR": (),
        "OBSERVABLE_INCLUDE": (),
        "ELSE_CORRELATED_ERROR": thrower("Converting ELSE_CORRELATED_ERROR to cirq is not supported."),
        "TICK": (),
    }


def not_handled_or_handled_specially_set() -> Set[str]:
    return {"E", "CORRELATED_ERROR", "REPEAT", "ELSE_CORRELATED_ERROR", "QUBIT_COORDS", "SHIFT_COORDS"}


class CoordTracker:
    def __init__(self):
        self.qubit_coords: Dict[int, cirq.Qid] = {}
        self.origin: DefaultDict[float] = collections.defaultdict(float)


def _translate_flattened_operation(
        op: Tuple[str, List, Union[float, Iterable[float]]],
        coord_tracker: CoordTracker,
        get_next_measure_id: Callable[[], int]) -> Iterator[cirq.Operation]:
    name, targets, arg = op

    handler = stim_to_cirq_gate_table().get(name)

    if handler is not None:
        if isinstance(handler, cirq.Gate):
            gate = handler
        elif handler == ():
            return
        else:
            gate = handler(arg)
        for q in targets:
            if isinstance(q, tuple) and q[0] == "rec":
                raise NotImplementedError("Measurement record.")
        if name[0] == 'M':  # Measurement.
            for t in targets:
                if isinstance(t, int):
                    q = t
                    g = gate
                elif t[0] == "inv":
                    q = t[1]
                    g = cast(Union[MeasureAndOrReset, cirq.MeasurementGate], gate).with_bits_flipped(0)
                else:
                    raise NotImplementedError("Unrecognized measurement target.")
                key = str(get_next_measure_id())
                yield g.with_key(key).on(cirq.LineQubit(q))
        else:
            m = cirq.num_qubits(gate)
            for k in range(0, len(targets), m):
                yield gate(*[cirq.LineQubit(q) for q in targets[k:k+m]])
        return

    if name == "E":
        yield cirq.PauliString({cirq.LineQubit(q): k for k, q in targets}).with_probability(arg)
        return

    if name == "QUBIT_COORDS":
        args: List[float] = [arg] if isinstance(arg, float) else arg
        for k in range(len(args)):
            args[k] += coord_tracker.origin[k]
            if args[k] == int(args[k]):
                args[k] = int(args[k])
        for t in targets:
            if len(args) == 1:
                coord_tracker.qubit_coords[t] = cirq.LineQubit(*args)
            elif len(args) == 2:
                coord_tracker.qubit_coords[t] = cirq.GridQubit(*args)
        return

    if name == "SHIFT_COORDS":
        args: List[float] = [arg] if isinstance(arg, float) else arg
        for k, a in enumerate(args):
            coord_tracker.origin[k] += a
        return

    raise NotImplementedError(f"Unsupported gate: {name}")


def stim_circuit_to_cirq_circuit(circuit: stim.Circuit) -> cirq.Circuit:
    """Converts a stim circuit into an equivalent cirq circuit.

    Qubit indices are turned into cirq.LineQubit instances. Measurements are
    keyed by their ordering (e.g. the first measurement is keyed "0", the second
    is keyed "1", etc).

    Not all circuits can be converted:
        - ELSE_CORRELATED_ERROR instructions are not supported.

    Not all circuits can be converted with perfect 1:1 fidelity:
        - DETECTOR annotations are discarded.
        - OBSERVABLE_INCLUDE annotations are discarded.

    Args:
        circuit: The stim circuit to convert into a cirq circuit.

    Returns:
        The converted circuit.

    Examples:

        >>> import stimcirq
        >>> import stim
        >>> print(stimcirq.stim_circuit_to_cirq_circuit(stim.Circuit('''
        ...     H 0
        ...     CNOT 0 1
        ...     X_ERROR(0.25) 0
        ...     TICK
        ...     M !1 0
        ... ''')))
        0: ───H───@───X[prob=0.25]───M('1')────
                  │
        1: ───────X──────────────────!M('0')───
    """
    _next_measure_id = 0

    def get_next_measure_id() -> int:
        nonlocal _next_measure_id
        _next_measure_id += 1
        return _next_measure_id - 1

    coord_tracker = CoordTracker()

    full_circuit = cirq.Circuit()
    current_tick = cirq.Circuit()
    for op in circuit.flattened_operations():
        if op[0] == 'TICK':
            if len(current_tick):
                full_circuit += current_tick
                current_tick = cirq.Circuit()
            else:
                full_circuit += cirq.Moment()
        else:
            current_tick += _translate_flattened_operation(op, coord_tracker, get_next_measure_id)
    full_circuit += current_tick

    if coord_tracker.qubit_coords:
        remap: Dict[cirq.Qid, cirq.Qid] = {
            q: coord_tracker.qubit_coords.get(cast(cirq.LineQubit, q).x, q)
            for q in full_circuit.all_qubits()
        }
        # Only remap if there are no collisions.
        if len(set(remap.values())) == len(remap):
            full_circuit = full_circuit.transform_qubits(remap.__getitem__)

    return full_circuit



from typing import Dict, Tuple, Sequence, List, Union

import cirq
import numpy as np
import pytest
import stim

import stimcirq
from stimcirq._cirq_to_stim import (
    cirq_circuit_to_stim_data,
    gate_to_stim_append_func,
)


def solve_tableau(gate: cirq.Gate) -> Dict[cirq.PauliString, cirq.PauliString]:
    """Computes a stabilizer tableau for the given gate."""

    result = {}

    n = gate.num_qubits()
    qs = cirq.LineQubit.range(n)
    for inp in [g(q) for g in [cirq.X, cirq.Z] for q in qs]:
        # Use superdense coding to extract X and Z flips from the generator conjugated by the gate.
        c = cirq.Circuit(
            cirq.H.on_each(qs),
            [cirq.CNOT(q, q + n) for q in qs],
            gate(*qs) ** -1,
            inp,
            gate(*qs),
            [cirq.CNOT(q, q + n) for q in qs],
            cirq.H.on_each(qs),
            [cirq.measure(q, q + n, key=str(q)) for q in qs],
        )

        # Extract X/Y/Z data from sample result (which should be deterministic).
        s = cirq.Simulator().sample(c)
        out: cirq.PauliString = cirq.PauliString({q: "IXZY"[s[str(q)][0]] for q in qs})

        # Use phase kickback to determine the sign of the output stabilizer.
        sign = cirq.NamedQubit('a')
        c = cirq.Circuit(
            cirq.H(sign),
            inp.controlled_by(sign),
            gate(*qs),
            out.controlled_by(sign),
            cirq.H(sign),
            cirq.measure(sign, key='sign'),
        )
        if cirq.Simulator().sample(c)['sign'][0]:
            out *= -1

        result[inp] = out
    return result


def test_solve_tableau():
    a, b = cirq.LineQubit.range(2)
    assert solve_tableau(cirq.I) == {cirq.X(a): cirq.X(a), cirq.Z(a): cirq.Z(a)}
    assert solve_tableau(cirq.S) == {cirq.X(a): cirq.Y(a), cirq.Z(a): cirq.Z(a)}
    assert solve_tableau(cirq.S ** -1) == {cirq.X(a): -cirq.Y(a), cirq.Z(a): cirq.Z(a)}
    assert solve_tableau(cirq.H) == {cirq.X(a): cirq.Z(a), cirq.Z(a): cirq.X(a)}
    assert solve_tableau(
        cirq.SingleQubitCliffordGate.from_xz_map((cirq.Y, False), (cirq.X, False))
    ) == {cirq.X(a): cirq.Y(a), cirq.Z(a): cirq.X(a)}
    assert solve_tableau(cirq.CZ) == {
        cirq.X(a): cirq.X(a) * cirq.Z(b),
        cirq.Z(a): cirq.Z(a),
        cirq.X(b): cirq.Z(a) * cirq.X(b),
        cirq.Z(b): cirq.Z(b),
    }


def assert_unitary_gate_converts_correctly(gate: cirq.Gate):
    n = gate.num_qubits()
    for pre, post in solve_tableau(gate).items():
        # Create a circuit that measures pre before the gate times post after the gate.
        # If the gate is translated correctly, the measurement will always be zero.

        c = stim.Circuit()
        c.append_operation("H", range(n))
        for i in range(n):
            c.append_operation("CNOT", [i, i + n])
        c.append_operation("H", [2 * n])
        for q, p in pre.items():
            c.append_operation(f"C{p}", [2 * n, q.x])
        qs = cirq.LineQubit.range(n)
        conv_gate, _ = cirq_circuit_to_stim_data(cirq.Circuit(gate(*qs)), q2i={q: q.x for q in qs})
        c += conv_gate
        for q, p in post.items():
            c.append_operation(f"C{p}", [2 * n, q.x])
        if post.coefficient == -1:
            c.append_operation("Z", [2 * n])
        c.append_operation("H", [2 * n])
        c.append_operation("M", [2 * n])
        correct = np.count_nonzero(c.compile_sampler().sample_bit_packed(10)) == 0
        assert correct, f"{gate!r} failed to turn {pre} into {post}.\nConverted to:\n{conv_gate}\n"


@pytest.mark.parametrize("gate", gate_to_stim_append_func().keys())
def test_unitary_gate_conversions(gate: cirq.Gate):
    # Note: filtering in the parametrize annotation causes a false 'gate undefined' lint error.
    if cirq.has_unitary(gate):
        assert_unitary_gate_converts_correctly(gate)


def test_more_unitary_gate_conversions():
    for p in [1, 1j, -1, -1j]:
        assert_unitary_gate_converts_correctly(p * cirq.DensePauliString("IXYZ"))
        assert_unitary_gate_converts_correctly((p * cirq.DensePauliString("IXYZ")).controlled(1))

    a, b = cirq.LineQubit.range(2)
    c, _ = cirq_circuit_to_stim_data(cirq.Circuit(cirq.H(a), cirq.CNOT(a, b), cirq.measure(a, b), cirq.reset(a)))
    assert (
        str(c).strip()
        == """
H 0
TICK
CX 0 1
TICK
M 0 1
TICK
R 0
TICK
    """.strip()
    )


ROUND_TRIP_NOISY_GATES = [
    cirq.BitFlipChannel(0.1),
    cirq.BitFlipChannel(0.2),
    cirq.PhaseFlipChannel(0.1),
    cirq.PhaseFlipChannel(0.2),
    cirq.PhaseDampingChannel(0.1),
    cirq.PhaseDampingChannel(0.2),
    cirq.X.with_probability(0.1),
    cirq.X.with_probability(0.2),
    cirq.Y.with_probability(0.1),
    cirq.Y.with_probability(0.2),
    cirq.Z.with_probability(0.1),
    cirq.Z.with_probability(0.2),
    cirq.DepolarizingChannel(0.1),
    cirq.DepolarizingChannel(0.2),
    cirq.DepolarizingChannel(0.1, n_qubits=2),
    cirq.DepolarizingChannel(0.2, n_qubits=2),
    cirq.AsymmetricDepolarizingChannel(p_x=0, p_y=0, p_z=0),
    cirq.AsymmetricDepolarizingChannel(p_x=0.2, p_y=0.1, p_z=0.3),
    cirq.AsymmetricDepolarizingChannel(p_x=0.1, p_y=0, p_z=0),
    cirq.AsymmetricDepolarizingChannel(p_x=0, p_y=0.1, p_z=0),
    cirq.AsymmetricDepolarizingChannel(p_x=0, p_y=0, p_z=0.1),
    *[
        stimcirq.TwoQubitAsymmetricDepolarizingChannel(cirq.one_hot(index=k, shape=15, value=0.1, dtype=np.float64))
        for k in range(15)
    ],
    stimcirq.TwoQubitAsymmetricDepolarizingChannel([k/300 for k in range(1, 16)]),
    stimcirq.TwoQubitAsymmetricDepolarizingChannel([0]*15),
]


@pytest.mark.parametrize("gate", ROUND_TRIP_NOISY_GATES)
def test_frame_simulator_sampling_noisy_gates_agrees_with_cirq_data(gate: cirq.Gate):
    # Create test circuit that uses superdense coding to quantify arbitrary Pauli error mixtures.
    n = cirq.num_qubits(gate)
    qs = cirq.LineQubit.range(n)
    circuit = cirq.Circuit(
        cirq.H.on_each(qs),
        [cirq.CNOT(q, q + n) for q in qs],
        gate(*qs),
        [cirq.CNOT(q, q + n) for q in qs],
        cirq.H.on_each(qs),
    )
    expected_rates = cirq.final_density_matrix(circuit).diagonal().real

    # Convert test circuit to Stim and sample from it.
    stim_circuit, _ = cirq_circuit_to_stim_data(circuit + cirq.measure(*sorted(circuit.all_qubits())[::-1]))
    sample_count = 10000
    samples = stim_circuit.compile_sampler().sample_bit_packed(sample_count).flat
    unique, counts = np.unique(samples, return_counts=True)

    # Compare sample rates to expected rates.
    for value, count in zip(unique, counts):
        expected_rate = expected_rates[value]
        actual_rate = count / sample_count
        allowed_variation = 5 * (expected_rate * (1 - expected_rate) / sample_count) ** 0.5
        if not 0 <= expected_rate - allowed_variation <= 1:
            raise ValueError("Not enough samples to bound results away from extremes.")
        assert abs(expected_rate - actual_rate) < allowed_variation, (
            f"Sample rate {actual_rate} is over 5 standard deviations away from {expected_rate}.\n"
            f"Gate: {gate}\n"
            f"Test circuit:\n{circuit}\n"
            f"Converted circuit:\n{stim_circuit}\n"
        )


@pytest.mark.parametrize("gate", ROUND_TRIP_NOISY_GATES)
def test_tableau_simulator_sampling_noisy_gates_agrees_with_cirq_data(gate: cirq.Gate):
    # Technically this be a test of the `stim` package itself, but it's so convenient to compare to cirq.

    # Create test circuit that uses superdense coding to quantify arbitrary Pauli error mixtures.
    n = cirq.num_qubits(gate)
    qs = cirq.LineQubit.range(n)
    circuit = cirq.Circuit(
        cirq.H.on_each(qs),
        [cirq.CNOT(q, q + n) for q in qs],
        gate(*qs),
        [cirq.CNOT(q, q + n) for q in qs],
        cirq.H.on_each(qs),
    )
    expected_rates = cirq.final_density_matrix(circuit).diagonal().real

    # Convert test circuit to Stim and sample from it.
    stim_circuit, _ = cirq_circuit_to_stim_data(circuit + cirq.measure(*sorted(circuit.all_qubits())[::-1]))
    sample_count = 10000
    samples = []
    for _ in range(sample_count):
        sim = stim.TableauSimulator()
        sim.do(stim_circuit)
        s = 0
        for k, v in enumerate(sim.current_measurement_record()):
            s |= v << k
        samples.append(s)

    unique, counts = np.unique(samples, return_counts=True)

    # Compare sample rates to expected rates.
    for value, count in zip(unique, counts):
        expected_rate = expected_rates[value]
        actual_rate = count / sample_count
        allowed_variation = 5 * (expected_rate * (1 - expected_rate) / sample_count) ** 0.5
        if not 0 <= expected_rate - allowed_variation <= 1:
            raise ValueError("Not enough samples to bound results away from extremes.")
        assert abs(expected_rate - actual_rate) < allowed_variation, (
            f"Sample rate {actual_rate} is over 5 standard deviations away from {expected_rate}.\n"
            f"Gate: {gate}\n"
            f"Test circuit:\n{circuit}\n"
            f"Converted circuit:\n{stim_circuit}\n"
        )


def test_cirq_circuit_to_stim_circuit_custom_stim_method():
    class DetectorGate(cirq.Gate):
        def _num_qubits_(self):
            return 1

        def _measure_keys_(self):
            return "custom",

        def _stim_conversion_(self,
                              edit_circuit: stim.Circuit,
                              edit_measurement_key_lengths: List[Tuple[str, int]],
                              targets: Sequence[int],
                              **kwargs):
            edit_measurement_key_lengths.append(("custom", 2))
            edit_circuit.append_operation("M", [stim.target_inv(targets[0])])
            edit_circuit.append_operation("M", [targets[0]])
            edit_circuit.append_operation("DETECTOR", [stim.target_rec(-1)])

    class SecondLastMeasurementWasDeterministicOperation(cirq.Operation):
        def _stim_conversion_(self, edit_circuit: stim.Circuit, **kwargs):
            edit_circuit.append_operation("DETECTOR", [stim.target_rec(-2)])

        def with_qubits(self, *new_qubits):
            raise NotImplementedError()

        @property
        def qubits(self) -> Tuple['cirq.Qid', ...]:
            return ()

    a, b, c = cirq.LineQubit.range(3)
    cirq_circuit = cirq.Circuit(
        cirq.measure(a, key="a"),
        cirq.measure(b, key="b"),
        cirq.measure(c, key="c"),
        cirq.Moment(SecondLastMeasurementWasDeterministicOperation()),
        cirq.Moment(DetectorGate().on(b)),
    )

    stim_circuit = stimcirq.cirq_circuit_to_stim_circuit(cirq_circuit)
    assert str(stim_circuit).strip() == """
M 0 1 2
TICK
DETECTOR rec[-2]
TICK
M !1 1
DETECTOR rec[-1]
TICK
    """.strip()

    class BadGate(cirq.Gate):
        def num_qubits(self) -> int:
            return 1

        def _stim_conversion_(self):
            pass

    with pytest.raises(TypeError, match="dont_forget_your_star_star_kwargs"):
        stimcirq.cirq_circuit_to_stim_circuit(cirq.Circuit(BadGate().on(a)))

    sample = stimcirq.StimSampler().sample(cirq_circuit)
    assert len(sample.columns) == 4
    np.testing.assert_array_equal(sample["a"], [0])
    np.testing.assert_array_equal(sample["b"], [0])
    np.testing.assert_array_equal(sample["c"], [0])
    np.testing.assert_array_equal(sample["custom"], [2])


def test_custom_qubit_indexing():
    a = cirq.NamedQubit("a")
    b = cirq.NamedQubit("b")
    actual = stimcirq.cirq_circuit_to_stim_circuit(cirq.Circuit(cirq.CNOT(a, b)), qubit_to_index_dict={a: 10, b: 15})
    assert actual == stim.Circuit('CX 10 15\nTICK')
